from uuid import uuid4

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_workspace():
    client = boto3.client("amp", region_name="ap-southeast-1")
    resp = client.create_workspace(alias="test", clientToken="mytoken")

    assert "arn" in resp
    assert resp["status"] == {"statusCode": "ACTIVE"}
    assert "workspaceId" in resp


@mock_aws
def test_describe_workspace():
    client = boto3.client("amp", region_name="eu-west-1")
    workspace_id = client.create_workspace(alias="test", clientToken="mytoken")[
        "workspaceId"
    ]

    resp = client.describe_workspace(workspaceId=workspace_id)
    assert "workspace" in resp

    workspace = resp["workspace"]
    assert "alias" in workspace
    assert "arn" in workspace
    assert "createdAt" in workspace
    assert "prometheusEndpoint" in workspace
    assert workspace["status"] == {"statusCode": "ACTIVE"}
    assert workspace["workspaceId"] == workspace_id


@mock_aws
def test_list_workspaces():
    my_alias = str(uuid4())[0:6]
    client = boto3.client("amp", region_name="ap-southeast-1")
    client.create_workspace(alias="test")
    client.create_workspace(alias=my_alias)

    spaces = client.list_workspaces(maxResults=1000)["workspaces"]
    assert len(spaces) >= 2
    assert "test" in [sp.get("alias") for sp in spaces]
    assert my_alias in [sp.get("alias") for sp in spaces]

    resp = client.list_workspaces(alias=my_alias)
    assert len(resp["workspaces"]) == 1
    assert resp["workspaces"][0]["alias"] == my_alias


@mock_aws
def test_list_workspaces__paginated():
    client = boto3.client("amp", region_name="ap-southeast-1")
    for _ in range(125):
        client.create_workspace()

    # default pagesize is 100
    page1 = client.list_workspaces()
    assert len(page1["workspaces"]) == 100
    assert "nextToken" in page1

    # We can ask for a smaller pagesize
    page2 = client.list_workspaces(maxResults=15, nextToken=page1["nextToken"])
    assert len(page2["workspaces"]) == 15
    assert "nextToken" in page2

    # We could request all of them in one go
    all_workspaces = client.list_workspaces(maxResults=1000)["workspaces"]
    length = len(all_workspaces)
    # We don't know exactly how much workspaces there are, because we are running multiple tests at the same time
    assert length >= 125


@mock_aws
def test_list_tags_for_resource():
    client = boto3.client("amp", region_name="ap-southeast-1")
    arn = client.create_workspace(
        alias="test", clientToken="mytoken", tags={"t1": "v1", "t2": "v2"}
    )["arn"]

    assert get_tags(arn, client) == {"t1": "v1", "t2": "v2"}


@mock_aws
def test_update_workspace_alias():
    client = boto3.client("amp", region_name="ap-southeast-1")

    workspace_id = client.create_workspace(alias="initial")["workspaceId"]

    w = client.describe_workspace(workspaceId=workspace_id)["workspace"]
    assert w["alias"] == "initial"

    client.update_workspace_alias(alias="updated", workspaceId=workspace_id)

    w = client.describe_workspace(workspaceId=workspace_id)["workspace"]
    assert w["alias"] == "updated"


@mock_aws
def test_delete_workspace():
    client = boto3.client("amp", region_name="us-east-2")

    workspace_id = client.create_workspace(alias="test", clientToken="mytoken")[
        "workspaceId"
    ]

    client.delete_workspace(workspaceId=workspace_id)

    with pytest.raises(ClientError) as exc:
        client.describe_workspace(workspaceId=workspace_id)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == "Workspace not found"


@mock_aws
def test_tag_resource():
    client = boto3.client("amp", region_name="us-east-2")

    workspace = client.create_workspace(alias="test", tags={"t": "v"})
    arn = workspace["arn"]
    workspace_id = workspace["workspaceId"]

    client.tag_resource(resourceArn=arn, tags={"t1": "v1", "t2": "v2"})

    expected = {"t": "v", "t1": "v1", "t2": "v2"}
    assert get_tags(arn, client) == expected
    assert (
        client.describe_workspace(workspaceId=workspace_id)["workspace"]["tags"]
        == expected
    )

    client.untag_resource(resourceArn=arn, tagKeys=["t1"])
    assert get_tags(arn, client) == {"t": "v", "t2": "v2"}

    client.untag_resource(resourceArn=arn, tagKeys=["t", "t2"])
    assert get_tags(arn, client) == {}


def get_tags(arn, client):
    return client.list_tags_for_resource(resourceArn=arn)["tags"]



import enum
from typing import Final

from moto.stepfunctions.parser.asl.component.component import Component


class CSVHeaderLocationValue(enum.Enum):
    FIRST_ROW = "FIRST_ROW"
    GIVEN = "GIVEN"


class CSVHeaderLocation(Component):
    csv_header_location_value: Final[CSVHeaderLocationValue]

    def __init__(self, csv_header_location_value: str):
        self.csv_header_location_value = CSVHeaderLocationValue(
            csv_header_location_value
        )  # Pass error upstream.



from datetime import datetime
from unittest import SkipTest

import boto3
from freezegun import freeze_time

from moto import mock_aws, settings


@mock_aws
def test_rfc_returns_valid_date_for_every_possible_weekday_and_month():
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest("Freezing time only possible in DecoratorMode")
    client = boto3.client("s3", region_name="us-east-1")
    client.create_bucket(Bucket="bucket_")
    for weekday in range(1, 8):
        with freeze_time(f"2023-02-{weekday} 12:00:00"):
            client.put_object(Bucket="bucket_", Key="test.txt", Body=b"test")
            obj = client.get_object(Bucket="bucket_", Key="test.txt")
            assert obj["LastModified"].replace(tzinfo=None) == datetime.now()
            # If we get here, the LastModified date will have been successfully parsed
            # Regardless of which weekday it is

    for month in range(1, 13):
        with freeze_time(f"2023-{month}-02 12:00:00"):
            client.put_object(Bucket="bucket_", Key="test.txt", Body=b"test")
            obj = client.get_object(Bucket="bucket_", Key="test.txt")
            assert obj["LastModified"].replace(tzinfo=None) == datetime.now()
            # If we get here, the LastModified date will have been successfully parsed
            # Regardless of which month it is



import json
from typing import Any, Dict, Tuple

from moto.core.responses import BaseResponse
from moto.events.models import EventsBackend, events_backends


class EventsHandler(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="events")

    @property
    def events_backend(self) -> EventsBackend:
        return events_backends[self.current_account][self.region]

    def _create_response(self, result: Any) -> Tuple[str, Dict[str, Any]]:
        """
        Creates a proper response for the API.

        It basically transforms a dict-like result from the backend
        into a tuple (str, dict) properly formatted.
        Args:
            result (dict): result from backend

        Returns:
            (str, dict): dumped result and headers
        """
        return json.dumps(result), self.response_headers

    def error(
        self, type_: str, message: str = "", status: int = 400
    ) -> Tuple[str, Dict[str, Any]]:
        headers: Dict[str, Any] = self.response_headers
        headers["status"] = status
        return json.dumps({"__type": type_, "message": message}), headers

    def put_rule(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        event_pattern = self._get_param("EventPattern")
        scheduled_expression = self._get_param("ScheduleExpression")
        state = self._get_param("State")
        desc = self._get_param("Description")
        role_arn = self._get_param("RoleArn")
        event_bus_arn = self._get_param("EventBusName")
        tags = self._get_param("Tags")

        rule = self.events_backend.put_rule(
            name,
            scheduled_expression=scheduled_expression,
            event_pattern=event_pattern,
            state=state,
            description=desc,
            role_arn=role_arn,
            event_bus_arn=event_bus_arn,
            tags=tags,
        )
        result = {"RuleArn": rule.arn}
        return self._create_response(result)

    def delete_rule(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        event_bus_arn = self._get_param("EventBusName")

        if not name:
            return self.error("ValidationException", "Parameter Name is required.")
        self.events_backend.delete_rule(name, event_bus_arn)

        return "", self.response_headers

    def describe_rule(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        event_bus_arn = self._get_param("EventBusName")

        if not name:
            return self.error("ValidationException", "Parameter Name is required.")

        rule = self.events_backend.describe_rule(name, event_bus_arn)

        result = rule.describe()
        return self._create_response(result)

    def disable_rule(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        event_bus_arn = self._get_param("EventBusName")

        if not name:
            return self.error("ValidationException", "Parameter Name is required.")

        if not self.events_backend.disable_rule(name, event_bus_arn):
            return self.error(
                "ResourceNotFoundException", "Rule " + name + " does not exist."
            )

        return "", self.response_headers

    def enable_rule(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        event_bus_arn = self._get_param("EventBusName")

        if not name:
            return self.error("ValidationException", "Parameter Name is required.")

        if not self.events_backend.enable_rule(name, event_bus_arn):
            return self.error(
                "ResourceNotFoundException", "Rule " + name + " does not exist."
            )

        return "", self.response_headers

    def generate_presigned_url(self) -> None:
        pass

    def list_rule_names_by_target(self) -> Tuple[str, Dict[str, Any]]:
        target_arn = self._get_param("TargetArn")
        event_bus_arn = self._get_param("EventBusName")
        next_token = self._get_param("NextToken")
        limit = self._get_param("Limit")

        if not target_arn:
            return self.error("ValidationException", "Parameter TargetArn is required.")

        rules, token = self.events_backend.list_rule_names_by_target(
            target_arn=target_arn,
            event_bus_arn=event_bus_arn,
            next_token=next_token,
            limit=limit,
        )

        res = {"RuleNames": [rule.name for rule in rules], "NextToken": token}

        return json.dumps(res), self.response_headers

    def list_rules(self) -> Tuple[str, Dict[str, Any]]:
        prefix = self._get_param("NamePrefix")
        event_bus_arn = self._get_param("EventBusName")
        next_token = self._get_param("NextToken")
        limit = self._get_param("Limit")

        rules, token = self.events_backend.list_rules(
            prefix=prefix,
            event_bus_arn=event_bus_arn,
            next_token=next_token,
            limit=limit,
        )
        rules_obj = {
            "Rules": [rule.describe() for rule in rules],
            "NextToken": token,
        }

        return json.dumps(rules_obj), self.response_headers

    def list_targets_by_rule(self) -> Tuple[str, Dict[str, Any]]:
        rule_name = self._get_param("Rule")
        event_bus_arn = self._get_param("EventBusName")
        next_token = self._get_param("NextToken")
        limit = self._get_param("Limit")

        if not rule_name:
            return self.error("ValidationException", "Parameter Rule is required.")

        try:
            targets = self.events_backend.list_targets_by_rule(
                rule_name, event_bus_arn, next_token, limit
            )
        except KeyError:
            return self.error(
                "ResourceNotFoundException", "Rule " + rule_name + " does not exist."
            )

        return json.dumps(targets), self.response_headers

    def put_events(self) -> str:
        events = self._get_param("Entries")

        entries = self.events_backend.put_events(events)

        failed_count = len([e for e in entries if "ErrorCode" in e])
        response = {
            "FailedEntryCount": failed_count,
            "Entries": entries,
        }

        return json.dumps(response)

    def put_targets(self) -> Tuple[str, Dict[str, Any]]:
        rule_name = self._get_param("Rule")
        event_bus_name = self._get_param("EventBusName")
        targets = self._get_param("Targets")

        self.events_backend.put_targets(rule_name, event_bus_name, targets)

        return (
            json.dumps({"FailedEntryCount": 0, "FailedEntries": []}),
            self.response_headers,
        )

    def remove_targets(self) -> Tuple[str, Dict[str, Any]]:
        rule_name = self._get_param("Rule")
        event_bus_name = self._get_param("EventBusName")
        ids = self._get_param("Ids")

        self.events_backend.remove_targets(rule_name, event_bus_name, ids)

        return (
            json.dumps({"FailedEntryCount": 0, "FailedEntries": []}),
            self.response_headers,
        )

    def test_event_pattern(self) -> None:
        pass

    def put_permission(self) -> str:
        event_bus_name = self._get_param("EventBusName")
        action = self._get_param("Action")
        principal = self._get_param("Principal")
        statement_id = self._get_param("StatementId")
        policy = self._get_param("Policy")
        condition = self._get_param("Condition")

        self.events_backend.put_permission(
            event_bus_name, action, principal, statement_id, condition, policy
        )

        return ""

    def remove_permission(self) -> str:
        event_bus_name = self._get_param("EventBusName")
        statement_id = self._get_param("StatementId")
        remove_all_permissions = self._get_param("RemoveAllPermissions")

        self.events_backend.remove_permission(
            event_bus_name, statement_id, remove_all_permissions
        )

        return ""

    def describe_event_bus(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")

        event_bus = self.events_backend.describe_event_bus(name)
        response = {"Name": event_bus.name, "Arn": event_bus.arn}

        if event_bus.policy:
            response["Policy"] = event_bus.policy

        return json.dumps(response), self.response_headers

    def create_event_bus(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        event_source_name = self._get_param("EventSourceName")
        tags = self._get_param("Tags")

        event_bus = self.events_backend.create_event_bus(name, event_source_name, tags)
        return json.dumps({"EventBusArn": event_bus.arn}), self.response_headers

    def list_event_buses(self) -> Tuple[str, Dict[str, Any]]:
        name_prefix = self._get_param("NamePrefix")
        # ToDo: add 'NextToken' & 'Limit' parameters

        response = []
        for event_bus in self.events_backend.list_event_buses(name_prefix):
            event_bus_response = {"Name": event_bus.name, "Arn": event_bus.arn}

            if event_bus.policy:
                event_bus_response["Policy"] = event_bus.policy

            response.append(event_bus_response)

        return json.dumps({"EventBuses": response}), self.response_headers

    def delete_event_bus(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")

        self.events_backend.delete_event_bus(name)

        return "", self.response_headers

    def list_tags_for_resource(self) -> Tuple[str, Dict[str, Any]]:
        arn = self._get_param("ResourceARN")

        result = self.events_backend.list_tags_for_resource(arn)

        return json.dumps(result), self.response_headers

    def tag_resource(self) -> Tuple[str, Dict[str, Any]]:
        arn = self._get_param("ResourceARN")
        tags = self._get_param("Tags")

        self.events_backend.tag_resource(arn, tags)

        return "{}", self.response_headers

    def untag_resource(self) -> Tuple[str, Dict[str, Any]]:
        arn = self._get_param("ResourceARN")
        tags = self._get_param("TagKeys")

        self.events_backend.untag_resource(arn, tags)

        return "{}", self.response_headers

    def create_archive(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("ArchiveName")
        source_arn = self._get_param("EventSourceArn")
        description = self._get_param("Description")
        event_pattern = self._get_param("EventPattern")
        retention = self._get_param("RetentionDays")

        archive = self.events_backend.create_archive(
            name, source_arn, description, event_pattern, retention
        )

        return (
            json.dumps(
                {
                    "ArchiveArn": archive.arn,
                    "CreationTime": archive.creation_time,
                    "State": archive.state,
                }
            ),
            self.response_headers,
        )

    def describe_archive(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("ArchiveName")

        result = self.events_backend.describe_archive(name)

        return json.dumps(result), self.response_headers

    def list_archives(self) -> Tuple[str, Dict[str, Any]]:
        name_prefix = self._get_param("NamePrefix")
        source_arn = self._get_param("EventSourceArn")
        state = self._get_param("State")

        result = self.events_backend.list_archives(name_prefix, source_arn, state)

        return json.dumps({"Archives": result}), self.response_headers

    def update_archive(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("ArchiveName")
        description = self._get_param("Description")
        event_pattern = self._get_param("EventPattern")
        retention = self._get_param("RetentionDays")

        result = self.events_backend.update_archive(
            name, description, event_pattern, retention
        )

        return json.dumps(result), self.response_headers

    def delete_archive(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("ArchiveName")

        self.events_backend.delete_archive(name)

        return "", self.response_headers

    def start_replay(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("ReplayName")
        description = self._get_param("Description")
        source_arn = self._get_param("EventSourceArn")
        start_time = self._get_param("EventStartTime")
        end_time = self._get_param("EventEndTime")
        destination = self._get_param("Destination")

        result = self.events_backend.start_replay(
            name, description, source_arn, start_time, end_time, destination
        )

        return json.dumps(result), self.response_headers

    def describe_replay(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("ReplayName")

        result = self.events_backend.describe_replay(name)

        return json.dumps(result), self.response_headers

    def list_replays(self) -> Tuple[str, Dict[str, Any]]:
        name_prefix = self._get_param("NamePrefix")
        source_arn = self._get_param("EventSourceArn")
        state = self._get_param("State")

        result = self.events_backend.list_replays(name_prefix, source_arn, state)

        return json.dumps({"Replays": result}), self.response_headers

    def cancel_replay(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("ReplayName")

        result = self.events_backend.cancel_replay(name)

        return json.dumps(result), self.response_headers

    def create_connection(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        description = self._get_param("Description")
        authorization_type = self._get_param("AuthorizationType")
        auth_parameters = self._get_param("AuthParameters")

        result = self.events_backend.create_connection(
            name, description, authorization_type, auth_parameters
        )

        return (
            json.dumps(
                {
                    "ConnectionArn": result.arn,
                    "ConnectionState": "AUTHORIZED",
                    "CreationTime": result.creation_time,
                    "LastModifiedTime": result.creation_time,
                }
            ),
            self.response_headers,
        )

    def list_connections(self) -> Tuple[str, Dict[str, Any]]:
        connections = self.events_backend.list_connections()
        result = []
        for connection in connections:
            result.append(
                {
                    "ConnectionArn": connection.arn,
                    "ConnectionState": "AUTHORIZED",
                    "CreationTime": connection.creation_time,
                    "LastModifiedTime": connection.creation_time,
                    "AuthorizationType": connection.authorization_type,
                }
            )

        return json.dumps({"Connections": result}), self.response_headers

    def describe_connection(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        result = self.events_backend.describe_connection(name)
        return json.dumps(result), self.response_headers

    def update_connection(self) -> Tuple[str, Dict[str, Any]]:
        updates = dict(
            name=self._get_param("Name"),
            description=self._get_param("Description"),
            authorization_type=self._get_param("AuthorizationType"),
            auth_parameters=self._get_param("AuthParameters"),
        )
        result = self.events_backend.update_connection(**updates)
        return self._create_response(result)

    def delete_connection(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        result = self.events_backend.delete_connection(name)
        return json.dumps(result), self.response_headers

    def create_api_destination(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        description = self._get_param("Description")
        connection_arn = self._get_param("ConnectionArn")
        invocation_endpoint = self._get_param("InvocationEndpoint")
        invocation_rate_limit_per_second = self._get_param(
            "InvocationRateLimitPerSecond"
        )
        http_method = self._get_param("HttpMethod")

        result = self.events_backend.create_api_destination(
            name,
            description,
            connection_arn,
            invocation_endpoint,
            invocation_rate_limit_per_second,
            http_method,
        )
        return self._create_response(result)

    def list_api_destinations(self) -> Tuple[str, Dict[str, Any]]:
        destinations = self.events_backend.list_api_destinations()
        result = []
        for destination in destinations:
            result.append(
                {
                    "ApiDestinationArn": destination.arn,
                    "Name": destination.name,
                    "ApiDestinationState": destination.state,
                    "ConnectionArn": destination.connection_arn,
                    "InvocationEndpoint": destination.invocation_endpoint,
                    "HttpMethod": destination.http_method,
                    "CreationTime": destination.creation_time,
                    "LastModifiedTime": destination.creation_time,
                }
            )

        return json.dumps({"ApiDestinations": result}), self.response_headers

    def describe_api_destination(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        result = self.events_backend.describe_api_destination(name)
        return self._create_response(result)

    def update_api_destination(self) -> Tuple[str, Dict[str, Any]]:
        updates = dict(
            connection_arn=self._get_param("ConnectionArn"),
            description=self._get_param("Description"),
            http_method=self._get_param("HttpMethod"),
            invocation_endpoint=self._get_param("InvocationEndpoint"),
            invocation_rate_limit_per_second=self._get_param(
                "InvocationRateLimitPerSecond"
            ),
            name=self._get_param("Name"),
        )

        result = self.events_backend.update_api_destination(**updates)
        return self._create_response(result)

    def delete_api_destination(self) -> Tuple[str, Dict[str, Any]]:
        name = self._get_param("Name")
        self.events_backend.delete_api_destination(name)
        return self._create_response({})

    def create_partner_event_source(self) -> str:
        name = self._get_param("Name")
        account_id = self._get_param("Account")
        self.events_backend.create_partner_event_source(
            name=name,
            account_id=account_id,
        )
        return "{}"

    def describe_event_source(self) -> str:
        name = self._get_param("Name")
        event_source = self.events_backend.describe_event_source(name)
        return json.dumps(event_source.to_dict())

    def describe_partner_event_source(self) -> str:
        name = self._get_param("Name")
        event_source = self.events_backend.describe_partner_event_source(name)
        return json.dumps({"Arn": event_source.arn, "Name": event_source.name})

    def delete_partner_event_source(self) -> str:
        name = self._get_param("Name")
        account_id = self._get_param("Account")
        self.events_backend.delete_partner_event_source(name, account_id)
        return "{}"

    def put_partner_events(self) -> str:
        entries = self._get_param("Entries")
        self.events_backend.put_partner_events(entries)
        return json.dumps({"Entries": [], "FailedEntryCount": 0})



from moto.backends import get_service_from_url


def test_get_service_from_url() -> None:
    assert get_service_from_url("https://s3.amazonaws.com") == "s3"
    assert get_service_from_url("https://bucket.s3.amazonaws.com") == "s3"
    assert get_service_from_url("https://unknown.com") is None



import boto3

from moto import mock_aws, settings
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

REGION = "us-east-1"


if not settings.TEST_SERVER_MODE:

    @mock_aws
    def test_pab_are_kept_separate():
        client = boto3.client("s3control", region_name=REGION)
        s3_client = boto3.client("s3", region_name=REGION)
        s3_client.create_bucket(Bucket="bucket")

        client.put_public_access_block(
            AccountId=ACCOUNT_ID,
            PublicAccessBlockConfiguration={
                "BlockPublicAcls": True,
                "IgnorePublicAcls": True,
                "BlockPublicPolicy": True,
                "RestrictPublicBuckets": True,
            },
        )

        s3_client.put_public_access_block(
            Bucket="bucket",
            PublicAccessBlockConfiguration={
                "BlockPublicAcls": True,
                "IgnorePublicAcls": False,
                "BlockPublicPolicy": True,
                "RestrictPublicBuckets": False,
            },
        )

        pab_from_control = client.get_public_access_block(AccountId=ACCOUNT_ID)
        assert pab_from_control["PublicAccessBlockConfiguration"] == {
            "BlockPublicAcls": True,
            "IgnorePublicAcls": True,
            "BlockPublicPolicy": True,
            "RestrictPublicBuckets": True,
        }

        pab_from_s3 = s3_client.get_public_access_block(Bucket="bucket")
        assert pab_from_s3["PublicAccessBlockConfiguration"] == {
            "BlockPublicAcls": True,
            "IgnorePublicAcls": False,
            "BlockPublicPolicy": True,
            "RestrictPublicBuckets": False,
        }

    @mock_aws
    def test_pab_are_kept_separate_with_inverse_mocks():
        client = boto3.client("s3control", region_name=REGION)
        s3_client = boto3.client("s3", region_name=REGION)
        s3_client.create_bucket(Bucket="bucket")

        client.put_public_access_block(
            AccountId=ACCOUNT_ID,
            PublicAccessBlockConfiguration={
                "BlockPublicAcls": True,
                "IgnorePublicAcls": True,
                "BlockPublicPolicy": True,
                "RestrictPublicBuckets": True,
            },
        )

        s3_client.put_public_access_block(
            Bucket="bucket",
            PublicAccessBlockConfiguration={
                "BlockPublicAcls": True,
                "IgnorePublicAcls": False,
                "BlockPublicPolicy": True,
                "RestrictPublicBuckets": False,
            },
        )

        pab_from_control = client.get_public_access_block(AccountId=ACCOUNT_ID)
        assert pab_from_control["PublicAccessBlockConfiguration"] == {
            "BlockPublicAcls": True,
            "IgnorePublicAcls": True,
            "BlockPublicPolicy": True,
            "RestrictPublicBuckets": True,
        }

        pab_from_s3 = s3_client.get_public_access_block(Bucket="bucket")
        assert pab_from_s3["PublicAccessBlockConfiguration"] == {
            "BlockPublicAcls": True,
            "IgnorePublicAcls": False,
            "BlockPublicPolicy": True,
            "RestrictPublicBuckets": False,
        }

    @mock_aws
    def test_access_point_read_write():
        # Setup
        bucket = "test-bucket"
        ap_client = boto3.client("s3control", region_name=REGION)
        s3_client = boto3.client("s3", region_name=REGION)
        s3_client.create_bucket(Bucket=bucket)

        read_ap = ap_client.create_access_point(
            AccountId=ACCOUNT_ID, Name="read-ap", Bucket=bucket
        )
        write_ap = ap_client.create_access_point(
            AccountId=ACCOUNT_ID, Name="write-ap", Bucket=bucket
        )

        content = b"This is test content"
        key = "test/object.txt"

        # Execute
        s3_client.put_object(
            Bucket=write_ap["AccessPointArn"],
            Key=key,
            Body=content,
            ContentType="text/plain",
        )

        # Verify
        assert (
            s3_client.get_object(Bucket=read_ap["AccessPointArn"], Key=key)[
                "Body"
            ].read()
            == content
        )
        assert s3_client.get_object(Bucket=bucket, Key=key)["Body"].read() == content



"""bedrockagent base URL and path."""

from .responses import AgentsforBedrockResponse

url_bases = [
    r"https?://bedrock-agent\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/.*$": AgentsforBedrockResponse.dispatch,
}



from typing import Optional

from jinja2 import Template

from moto.core.exceptions import RESTError


class UnformattedGetAttTemplateException(Exception):
    description = (
        "Template error: resource {0} does not support attribute type {1} in Fn::GetAtt"
    )
    status_code = 400


class ValidationError(RESTError):
    def __init__(self, name_or_id: Optional[str] = None, message: Optional[str] = None):
        if message is None:
            message = f"Stack with id {name_or_id} does not exist"

        template = Template(ERROR_RESPONSE)
        super().__init__(error_type="ValidationError", message=message)
        self.description = template.render(code="ValidationError", message=message)


class MissingParameterError(RESTError):
    def __init__(self, parameter_name: str):
        template = Template(ERROR_RESPONSE)
        message = f"Missing parameter {parameter_name}"
        super().__init__(error_type="ValidationError", message=message)
        self.description = template.render(code="Missing Parameter", message=message)


class ExportNotFound(RESTError):
    """Exception to raise if a template tries to import a non-existent export"""

    def __init__(self, export_name: str):
        template = Template(ERROR_RESPONSE)
        message = f"No export named {export_name} found."
        super().__init__(error_type="ExportNotFound", message=message)
        self.description = template.render(code="ExportNotFound", message=message)


class StackSetNotEmpty(RESTError):
    def __init__(self) -> None:
        template = Template(ERROR_RESPONSE)
        message = "StackSet is not empty"
        super().__init__(error_type="StackSetNotEmptyException", message=message)
        self.description = template.render(
            code="StackSetNotEmptyException", message=message
        )


class StackSetNotFoundException(RESTError):
    def __init__(self, name: str):
        template = Template(ERROR_RESPONSE)
        message = f"StackSet {name} not found"
        super().__init__(error_type="StackSetNotFoundException", message=message)
        self.description = template.render(
            code="StackSetNotFoundException", message=message
        )


class UnsupportedAttribute(ValidationError):
    def __init__(self, resource: str, attr: str):
        template = Template(ERROR_RESPONSE)
        super().__init__()
        self.description = template.render(
            code="ValidationError",
            message=f"Template error: resource {resource} does not support attribute type {attr} in Fn::GetAtt",
        )


ERROR_RESPONSE = """<ErrorResponse xmlns="http://cloudformation.amazonaws.com/doc/2010-05-15/">
  <Error>
    <Type>Sender</Type>
    <Code>{{ code }}</Code>
    <Message>{{ message }}</Message>
  </Error>
  <RequestId>cf4c737e-5ae2-11e4-a7c9-ad44eEXAMPLE</RequestId>
</ErrorResponse>
"""



from moto.dynamodb.models.utilities import find_nested_key


class TestFindDictionaryKeys:
    def setup_method(self):
        self.item = {
            "simplestring": "val",
            "nesteddict": {
                "level21": {"level3.1": "val", "level3.2": "val"},
                "level22": {"level3.1": "val", "level3.2": "val"},
                "nestedlist": [
                    {"ll21": {"ll3.1": "val", "ll3.2": "val"}},
                    {"ll22": {"ll3.1": "val", "ll3.2": "val"}},
                ],
            },
            "rootlist": [
                {"ll21": {"ll3.1": "val", "ll3.2": "val"}},
                {"ll22": {"ll3.1": "val", "ll3.2": "val"}},
            ],
        }

    def test_find_nothing(self):
        assert find_nested_key([""], self.item) == {}

    def test_find_unknown_key(self):
        assert find_nested_key(["unknown"], self.item) == {}

    def test_project_single_key_string(self):
        assert find_nested_key(["simplestring"], self.item) == {"simplestring": "val"}

    def test_project_single_key_dict(self):
        assert find_nested_key(["nesteddict"], self.item) == {
            "nesteddict": {
                "level21": {"level3.1": "val", "level3.2": "val"},
                "level22": {"level3.1": "val", "level3.2": "val"},
                "nestedlist": [
                    {"ll21": {"ll3.1": "val", "ll3.2": "val"}},
                    {"ll22": {"ll3.1": "val", "ll3.2": "val"}},
                ],
            }
        }

    def test_project_nested_key(self):
        assert find_nested_key(["nesteddict", "level21"], self.item) == {
            "nesteddict": {"level21": {"level3.1": "val", "level3.2": "val"}}
        }

    def test_project_multi_level_nested_key(self):
        assert find_nested_key(["nesteddict", "level21", "level3.2"], self.item) == {
            "nesteddict": {"level21": {"level3.2": "val"}}
        }

    def test_project_nested_key__partial_fix(self):
        assert find_nested_key(["nesteddict", "levelunknown"], self.item) == {}

    def test_project_nested_key__partial_fix2(self):
        assert find_nested_key(["nesteddict", "unknown", "unknown2"], self.item) == {}

    def test_list_index(self):
        assert find_nested_key(["rootlist[0]"], self.item) == {
            "rootlist": [{"ll21": {"ll3.1": "val", "ll3.2": "val"}}]
        }

    def test_nested_list_index(self):
        assert find_nested_key(["nesteddict", "nestedlist[1]"], self.item) == {
            "nesteddict": {"nestedlist": [{"ll22": {"ll3.1": "val", "ll3.2": "val"}}]}
        }

    def test_nested_obj_in_list(self):
        assert find_nested_key(
            ["nesteddict", "nestedlist[1]", "ll22", "ll3.1"], self.item
        ) == {"nesteddict": {"nestedlist": [{"ll22": {"ll3.1": "val"}}]}}

    def test_list_unknown_indexes(self):
        assert find_nested_key(["nesteddict", "nestedlist[25]"], self.item) == {}



import os
import string
from typing import Any, List
from urllib.parse import urlparse

import yaml

from moto.moto_api._internal import mock_random as random
from moto.utilities.utils import get_partition


def generate_stack_id(stack_name: str, region: str, account: str) -> str:
    random_id = random.uuid4()
    return f"arn:{get_partition(region)}:cloudformation:{region}:{account}:stack/{stack_name}/{random_id}"


def generate_changeset_id(
    changeset_name: str, region_name: str, account_id: str
) -> str:
    random_id = random.uuid4()
    return f"arn:{get_partition(region_name)}:cloudformation:{region_name}:{account_id}:changeSet/{changeset_name}/{random_id}"


def generate_stackset_id(stackset_name: str) -> str:
    random_id = random.uuid4()
    return f"{stackset_name}:{random_id}"


def generate_stackset_arn(stackset_id: str, region_name: str, account_id: str) -> str:
    return f"arn:{get_partition(region_name)}:cloudformation:{region_name}:{account_id}:stackset/{stackset_id}"


def random_suffix() -> str:
    size = 12
    chars = list(range(10)) + list(string.ascii_uppercase)
    return "".join(str(random.choice(chars)) for x in range(size))


def yaml_tag_constructor(loader: Any, tag: Any, node: Any) -> Any:
    """convert shorthand intrinsic function to full name"""

    def _f(loader: Any, tag: Any, node: Any) -> Any:
        if tag == "!GetAtt":
            if isinstance(node.value, list):
                return node.value
            return node.value.split(".")
        elif type(node) == yaml.SequenceNode:
            return loader.construct_sequence(node)
        else:
            return node.value

    if tag == "!Ref":
        key = "Ref"
    else:
        key = f"Fn::{tag[1:]}"

    return {key: _f(loader, tag, node)}


def validate_template_cfn_lint(template: str) -> List[Any]:
    # Importing cfnlint adds a significant overhead, so we keep it local

    try:
        # Compatibility for cfn-lint 0.x
        # Fail fast with `cfnlint.core.configure_logging` which is removed in cfn-lint 1.x

        from cfnlint.core import configure_logging, get_rules, run_checks
        from cfnlint.decode import decode

        # Save the template to a temporary file -- cfn-lint requires a file
        filename = "file.tmp"
        with open(filename, "w") as file:
            file.write(template)
        abs_filename = os.path.abspath(filename)

        # decode handles both yaml and json
        try:
            template, matches = decode(abs_filename, False)
        except TypeError:
            # As of cfn-lint 0.39.0, the second argument (ignore_bad_template) was dropped
            # https://github.com/aws-cloudformation/cfn-python-lint/pull/1580
            template, matches = decode(abs_filename)

        # Set cfn-lint to info
        configure_logging(None)

        # Initialize the ruleset to be applied (no overrules, no excludes)
        rules = get_rules([], [], [])

        # Use us-east-1 region (spec file) for validation
        regions = ["us-east-1"]

        # Process all the rules and gather the errors
        return run_checks(abs_filename, template, rules, regions)

    except ImportError:
        # Compatibility for cfn-lint 1.x

        from cfnlint.api import lint
        from cfnlint.config import configure_logging
        from cfnlint.core import get_rules

        # Set cfn-lint to info
        configure_logging(None, False)

        # Initialize the ruleset to be applied (no overrules, no excludes)
        rules = get_rules([], [], [])

        # Use us-east-1 region (spec file) for validation
        regions = ["us-east-1"]

        # Process all the rules and gather the errors
        return lint(template, rules, regions)


def get_stack_from_s3_url(template_url: str, account_id: str, partition: str) -> str:
    from moto.s3.models import s3_backends

    template_url_parts = urlparse(template_url)
    if "localhost" in template_url:
        bucket_name, key_name = template_url_parts.path.lstrip("/").split("/", 1)
    else:
        if template_url_parts.netloc.endswith(
            "amazonaws.com"
        ) and template_url_parts.netloc.startswith("s3"):
            # Handle when S3 url uses amazon url with bucket in path
            # Also handles getting region as technically s3 is region'd

            # region = template_url.netloc.split('.')[1]
            bucket_name, key_name = template_url_parts.path.lstrip("/").split("/", 1)
        else:
            bucket_name = template_url_parts.netloc.split(".")[0]
            key_name = template_url_parts.path.lstrip("/")

    key = s3_backends[account_id][partition].get_object(bucket_name, key_name)
    return key.value.decode("utf-8")  # type: ignore[union-attr]



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws


@mock_aws
def test_filter_with_primary_region():
    conn = boto3.client("secretsmanager", region_name="us-east-1")

    conn.create_secret(Name="simple", SecretString="secret")
    conn.create_secret(
        Name="dup", SecretString="s", AddReplicaRegions=[{"Region": "us-east-2"}]
    )

    secrets = conn.list_secrets(
        Filters=[{"Key": "primary-region", "Values": ["us-east-1"]}]
    )["SecretList"]
    assert len(secrets) == 1

    secrets = conn.list_secrets(
        Filters=[{"Key": "primary-region", "Values": ["us-east-2"]}]
    )["SecretList"]
    assert len(secrets) == 0

    secrets = conn.list_secrets()["SecretList"]
    assert len(secrets) == 2

    # If our entry point is us-east-2, but filter in us-east-1
    # We can see the replicas in us-east-2 that have Primary Secrets in us-east-1
    use2 = boto3.client("secretsmanager", region_name="us-east-2")
    secrets = use2.list_secrets(
        Filters=[{"Key": "primary-region", "Values": ["us-east-1"]}]
    )["SecretList"]
    assert len(secrets) == 1
    assert secrets[0]["PrimaryRegion"] == "us-east-1"


@mock_aws
def test_create_secret_that_duplicates():
    use1 = boto3.client("secretsmanager", region_name="us-east-1")
    use2 = boto3.client("secretsmanager", region_name="us-east-2")

    resp = use1.create_secret(
        Name="dup", SecretString="s", AddReplicaRegions=[{"Region": "us-east-2"}]
    )
    primary_arn = resp["ARN"]
    assert resp["ReplicationStatus"][0]["Region"] == "us-east-2"
    assert resp["ReplicationStatus"][0]["Status"] == "InSync"

    primary_secret = use1.describe_secret(SecretId=primary_arn)
    assert primary_secret["ARN"] == primary_arn

    replica_arn = primary_arn.replace("us-east-1", "us-east-2")
    replica_secret = use2.describe_secret(SecretId=replica_arn)
    assert replica_secret["ARN"] == replica_arn
    assert replica_secret["Name"] == "dup"
    assert replica_secret["VersionIdsToStages"] == primary_secret["VersionIdsToStages"]

    replica_value = use2.get_secret_value(SecretId=replica_arn)
    assert replica_value["Name"] == "dup"
    assert replica_value["SecretString"] == "s"


@mock_aws
def test_create_secret_that_duplicates__in_same_region():
    conn = boto3.client("secretsmanager", region_name="us-east-1")

    with pytest.raises(ClientError) as exc:
        conn.create_secret(
            Name="dup_same_region",
            SecretString="s",
            AddReplicaRegions=[{"Region": "us-east-1"}],
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == "Invalid replica region."


@mock_aws
def test_replicate_secret():
    us1 = boto3.client("secretsmanager", region_name="us-east-1")
    us2 = boto3.client("secretsmanager", region_name="us-east-2")

    resp = us1.create_secret(Name="dup", SecretString="s")
    primary_arn = resp["ARN"]

    resp = us1.replicate_secret_to_regions(
        SecretId=primary_arn, AddReplicaRegions=[{"Region": "us-east-2"}]
    )
    assert resp["ReplicationStatus"][0]["Region"] == "us-east-2"
    assert resp["ReplicationStatus"][0]["Status"] == "InSync"

    replica_arn = primary_arn.replace("us-east-1", "us-east-2")
    replica_secret = us2.describe_secret(SecretId=replica_arn)
    assert replica_secret["ARN"] == replica_arn


@mock_aws
def test_replicate_secret__in_same_region():
    conn = boto3.client("secretsmanager", region_name="us-east-1")

    resp = conn.create_secret(Name="dup", SecretString="s")
    with pytest.raises(ClientError) as exc:
        conn.replicate_secret_to_regions(
            SecretId=resp["ARN"], AddReplicaRegions=[{"Region": "us-east-1"}]
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == "Invalid replica region."


@mock_aws
def test_replicate_secret__existing_secret():
    use1 = boto3.client("secretsmanager", region_name="us-east-1")
    use2 = boto3.client("secretsmanager", region_name="us-east-2")

    secret1 = use1.create_secret(Name="dup", SecretString="s1")["ARN"]
    secret2 = use2.create_secret(Name="dup", SecretString="s2")["ARN"]

    resp = use1.replicate_secret_to_regions(
        SecretId=secret1, AddReplicaRegions=[{"Region": "us-east-2"}]
    )
    assert resp["ReplicationStatus"][0]["Region"] == "us-east-2"
    assert resp["ReplicationStatus"][0]["Status"] == "Failed"
    assert (
        resp["ReplicationStatus"][0]["StatusMessage"]
        == "Replication failed: Secret name simple already exists in region us-east-2."
    )

    # us-east-2 still only has one secret
    assert len(use2.list_secrets()["SecretList"]) == 1

    # Both secrets still have original values
    assert use1.get_secret_value(SecretId=secret1)["SecretString"] == "s1"
    assert use2.get_secret_value(SecretId=secret2)["SecretString"] == "s2"


@mock_aws
def test_replicate_secret__overwrite_existing_secret():
    use1 = boto3.client("secretsmanager", region_name="us-east-1")
    use2 = boto3.client("secretsmanager", region_name="us-east-2")

    secret1 = use1.create_secret(Name="dup", SecretString="s1")["ARN"]
    secret2 = use2.create_secret(Name="dup", SecretString="s2")["ARN"]

    resp = use1.replicate_secret_to_regions(
        SecretId=secret1,
        AddReplicaRegions=[{"Region": "us-east-2"}],
        ForceOverwriteReplicaSecret=True,
    )
    assert resp["ReplicationStatus"][0]["Status"] == "InSync"

    # us-east-2 still only has one secret
    assert len(use2.list_secrets()["SecretList"]) == 1

    # Original Secret was yeeted
    with pytest.raises(ClientError):
        assert use2.get_secret_value(SecretId=secret2)["SecretString"] == "s1"

    # Use ListSecrets to get the new ARN
    # And verify it returns a copy of the us-east-1 secret
    secret2 = use2.list_secrets()["SecretList"][0]["ARN"]
    assert use2.get_secret_value(SecretId=secret2)["SecretString"] == "s1"


@mock_aws
def test_remove_regions_from_replication():
    use1 = boto3.client("secretsmanager", region_name="us-east-1")
    use2 = boto3.client("secretsmanager", region_name="us-east-2")
    usw2 = boto3.client("secretsmanager", region_name="us-west-2")

    arn1 = use1.create_secret(Name="dup", SecretString="s1")["ARN"]
    arn2 = arn1.replace("us-east-1", "us-east-2")
    arn3 = arn1.replace("us-east-1", "us-west-2")

    # Replicate to multiple regions
    use1.replicate_secret_to_regions(
        SecretId=arn1,
        AddReplicaRegions=[{"Region": "us-east-2"}, {"Region": "us-west-2"}],
    )

    # Replications exist
    use2.describe_secret(SecretId=arn2)
    usw2.describe_secret(SecretId=arn3)

    # Primary Secret knows about replicas
    replications = use1.describe_secret(SecretId=arn1)["ReplicationStatus"]
    assert set([rep["Region"] for rep in replications]) == {"us-east-2", "us-west-2"}

    # Remove us-east-2 replication
    use1.remove_regions_from_replication(
        SecretId=arn1, RemoveReplicaRegions=["us-east-2"]
    )
    with pytest.raises(ClientError):
        use2.describe_secret(SecretId=arn2)

    # us-west still exists
    usw2.describe_secret(SecretId=arn3)

    # Primary Secret no longer knows about us-east-2
    replications = use1.describe_secret(SecretId=arn1)["ReplicationStatus"]
    assert set([rep["Region"] for rep in replications]) == {"us-west-2"}

    # Removing region is idempotent - invoking it again does not result in any errors
    use1.remove_regions_from_replication(
        SecretId=arn1, RemoveReplicaRegions=["us-east-2"]
    )


@mock_aws
def test_update_replica():
    use1 = boto3.client("secretsmanager", region_name="us-east-1")
    use2 = boto3.client("secretsmanager", region_name="us-east-2")

    arn1 = use1.create_secret(
        Name="dup", SecretString="s1", AddReplicaRegions=[{"Region": "us-east-2"}]
    )["ARN"]
    arn2 = arn1.replace("us-east-1", "us-east-2")

    OP_NOT_PERMITTED = "Operation not permitted on a replica secret. Call must be made in primary secret's region."

    with pytest.raises(ClientError) as exc:
        use2.update_secret(SecretId=arn2, SecretString="asdf")
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.put_secret_value(SecretId=arn2, SecretString="asdf")
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.tag_resource(SecretId=arn2, Tags=[{"Key": "k", "Value": "v"}])
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.untag_resource(SecretId=arn2, TagKeys=["k1"])
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.put_resource_policy(SecretId=arn2, ResourcePolicy="k1")
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.get_resource_policy(SecretId=arn2)
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.delete_resource_policy(SecretId=arn2)
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.delete_secret(SecretId=arn2)
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.cancel_rotate_secret(SecretId=arn2)
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.rotate_secret(SecretId=arn2)
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.restore_secret(SecretId=arn2)
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.replicate_secret_to_regions(SecretId=arn2, AddReplicaRegions=[{}])
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED

    with pytest.raises(ClientError) as exc:
        use2.remove_regions_from_replication(SecretId=arn2, RemoveReplicaRegions=["a"])
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert err["Message"] == OP_NOT_PERMITTED


@mock_aws
def test_delete_while_duplication_exist():
    use1 = boto3.client("secretsmanager", region_name="us-east-1")

    region = "us-east-2"
    arn = use1.create_secret(
        Name="dup", SecretString="s1", AddReplicaRegions=[{"Region": region}]
    )["ARN"]

    # unable to delete
    with pytest.raises(ClientError) as exc:
        use1.delete_secret(SecretId=arn)
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterException"
    assert (
        err["Message"]
        == f"You can't delete secret {arn} that still has replica regions [{region}]"
    )

    # Remove us-east-2 replication
    use1.remove_regions_from_replication(SecretId=arn, RemoveReplicaRegions=[region])

    # Now we can delete
    use1.delete_secret(SecretId=arn)



from moto.core.exceptions import AWSError


class ExecutionAlreadyExists(AWSError):
    TYPE = "ExecutionAlreadyExists"
    STATUS = 400


class ExecutionDoesNotExist(AWSError):
    TYPE = "ExecutionDoesNotExist"
    STATUS = 400


class InvalidArn(AWSError):
    TYPE = "InvalidArn"
    STATUS = 400


class InvalidName(AWSError):
    TYPE = "InvalidName"
    STATUS = 400


class InvalidExecutionInput(AWSError):
    TYPE = "InvalidExecutionInput"
    STATUS = 400


class StateMachineDoesNotExist(AWSError):
    TYPE = "StateMachineDoesNotExist"
    STATUS = 400


class InvalidToken(AWSError):
    TYPE = "InvalidToken"
    STATUS = 400

    def __init__(self, message: str = "Invalid token"):
        super().__init__(f"Invalid Token: {message}")


class ResourceNotFound(AWSError):
    TYPE = "ResourceNotFound"
    STATUS = 400

    def __init__(self, arn: str):
        super().__init__(f"Resource not found: '{arn}'")


class ValidationException(AWSError):
    TYPE = "ValidationException"

    def __init__(self, msg: str):
        super().__init__(msg)


class NameTooLongException(ValidationException):
    def __init__(self, name: str):
        super().__init__(
            f"1 validation error detected: Value '{name}' at 'name' "
            "failed to satisfy constraint: Member must have length less than or equal to 80"
        )



from moto.core.exceptions import JsonRESTError


class WAFv2ClientError(JsonRESTError):
    code = 400


class WAFV2DuplicateItemException(WAFv2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "WafV2DuplicateItem",
            "AWS WAF could not perform the operation because some resource in your request is a duplicate of an existing one.",
        )


class WAFNonexistentItemException(WAFv2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "WAFNonexistentItemException",
            "AWS WAF couldn’t perform the operation because your resource doesn’t exist.",
        )


class WAFOptimisticLockException(WAFv2ClientError):
    def __init__(self) -> None:
        super().__init__(
            "WAFOptimisticLockException",
            "AWS WAF couldn’t save your changes because someone changed the resource after you started to edit it. Reapply your changes.",
        )



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws


@mock_aws
def test_create_vpn_connections_boto3():
    client = boto3.client("ec2", region_name="us-east-1")
    vpn_connection = client.create_vpn_connection(
        Type="ipsec.1", VpnGatewayId="vgw-0123abcd", CustomerGatewayId="cgw-0123abcd"
    )["VpnConnection"]
    assert vpn_connection["VpnConnectionId"].startswith("vpn-")
    assert vpn_connection["Type"] == "ipsec.1"


@mock_aws
def test_delete_vpn_connections_boto3():
    client = boto3.client("ec2", region_name="us-east-1")
    vpn_connection = client.create_vpn_connection(
        Type="ipsec.1", VpnGatewayId="vgw-0123abcd", CustomerGatewayId="cgw-0123abcd"
    )["VpnConnection"]

    conns = retrieve_all_vpncs(client)
    assert vpn_connection["VpnConnectionId"] in [c["VpnConnectionId"] for c in conns]

    client.delete_vpn_connection(VpnConnectionId=vpn_connection["VpnConnectionId"])

    conns = retrieve_all_vpncs(client)
    assert vpn_connection["VpnConnectionId"] in [c["VpnConnectionId"] for c in conns]
    my_cnx = [
        c for c in conns if c["VpnConnectionId"] == vpn_connection["VpnConnectionId"]
    ][0]
    assert my_cnx["State"] == "deleted"


@mock_aws
def test_delete_vpn_connections_bad_id_boto3():
    client = boto3.client("ec2", region_name="us-east-1")
    with pytest.raises(ClientError) as ex:
        client.delete_vpn_connection(VpnConnectionId="vpn-0123abcd")
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert "RequestId" in ex.value.response["ResponseMetadata"]
    assert ex.value.response["Error"]["Code"] == "InvalidVpnConnectionID.NotFound"


@mock_aws
def test_create_vpn_connection_with_vpn_gateway():
    client = boto3.client("ec2", region_name="us-east-1")

    vpn_gateway = client.create_vpn_gateway(Type="ipsec.1").get("VpnGateway", {})
    customer_gateway = client.create_customer_gateway(
        Type="ipsec.1", PublicIp="205.251.242.54", BgpAsn=65534
    ).get("CustomerGateway", {})
    vpn_connection = client.create_vpn_connection(
        Type="ipsec.1",
        VpnGatewayId=vpn_gateway["VpnGatewayId"],
        CustomerGatewayId=customer_gateway["CustomerGatewayId"],
    ).get("VpnConnection", {})

    assert vpn_connection["Type"] == "ipsec.1"
    assert vpn_connection["VpnGatewayId"] == vpn_gateway["VpnGatewayId"]
    assert vpn_connection["CustomerGatewayId"] == customer_gateway["CustomerGatewayId"]


@mock_aws
def test_describe_vpn_connections_boto3():
    client = boto3.client("ec2", region_name="us-east-1")

    vpn_gateway = client.create_vpn_gateway(Type="ipsec.1").get("VpnGateway", {})
    customer_gateway = client.create_customer_gateway(
        Type="ipsec.1", PublicIp="205.251.242.54", BgpAsn=65534
    ).get("CustomerGateway", {})
    vpn_connection1 = client.create_vpn_connection(
        Type="ipsec.1",
        VpnGatewayId=vpn_gateway["VpnGatewayId"],
        CustomerGatewayId=customer_gateway["CustomerGatewayId"],
    )["VpnConnection"]
    vpn_connection2 = client.create_vpn_connection(
        Type="ipsec.1",
        VpnGatewayId=vpn_gateway["VpnGatewayId"],
        CustomerGatewayId=customer_gateway["CustomerGatewayId"],
    )["VpnConnection"]

    conns = retrieve_all_vpncs(client)
    assert vpn_connection1["VpnConnectionId"] in [c["VpnConnectionId"] for c in conns]
    assert vpn_connection2["VpnConnectionId"] in [c["VpnConnectionId"] for c in conns]

    conns = client.describe_vpn_connections(
        VpnConnectionIds=[vpn_connection2["VpnConnectionId"]]
    )["VpnConnections"]

    assert conns[0]["VpnConnectionId"] == vpn_connection2["VpnConnectionId"]
    assert conns[0]["VpnGatewayId"] == vpn_gateway["VpnGatewayId"]
    assert conns[0]["Type"] == "ipsec.1"
    assert conns[0]["CustomerGatewayId"] == customer_gateway["CustomerGatewayId"]
    assert conns[0]["State"] == "available"


@mock_aws
def test_describe_vpn_connections_unknown():
    client = boto3.client("ec2", region_name="us-east-1")

    with pytest.raises(ClientError) as ex:
        client.describe_vpn_connections(VpnConnectionIds=["?"])
    err = ex.value.response["Error"]
    assert err["Message"] == "The vpnConnection ID '?' does not exist"
    assert err["Code"] == "InvalidVpnConnectionID.NotFound"


def retrieve_all_vpncs(client, filters=[]):  # pylint: disable=W0102
    resp = client.describe_vpn_connections(Filters=filters)
    all_vpncs = resp["VpnConnections"]
    token = resp.get("NextToken")
    while token:
        resp = client.describe_vpn_connections(NextToken=token, Filters=filters)
        all_vpncs.extend(resp["VpnConnections"])
        token = resp.get("NextToken")
    return all_vpncs



import datetime
import re

from moto.core import DEFAULT_ACCOUNT_ID
from moto.organizations import utils


def test_make_random_org_id():
    org_id = utils.make_random_org_id()
    assert re.match(utils.ORG_ID_REGEX, org_id)


def test_make_random_root_id():
    root_id = utils.make_random_root_id()
    assert re.match(utils.ROOT_ID_REGEX, root_id)


def test_make_random_ou_id():
    root_id = utils.make_random_root_id()
    ou_id = utils.make_random_ou_id(root_id)
    assert re.match(utils.OU_ID_REGEX, ou_id)


def test_make_random_account_id():
    account_id = utils.make_random_account_id()
    assert re.match(utils.ACCOUNT_ID_REGEX, account_id)


def test_make_random_create_account_status_id():
    create_account_status_id = utils.make_random_create_account_status_id()
    assert re.match(utils.CREATE_ACCOUNT_STATUS_ID_REGEX, create_account_status_id)


def test_make_random_policy_id():
    policy_id = utils.make_random_policy_id()
    assert re.match(utils.POLICY_ID_REGEX, policy_id)


def validate_organization(response, partition="aws"):
    org = response["Organization"]
    assert sorted(org.keys()) == [
        "Arn",
        "AvailablePolicyTypes",
        "FeatureSet",
        "Id",
        "MasterAccountArn",
        "MasterAccountEmail",
        "MasterAccountId",
    ]
    assert re.match(utils.ORG_ID_REGEX, org["Id"])
    assert org["MasterAccountId"] == DEFAULT_ACCOUNT_ID
    assert org["MasterAccountArn"] == (
        utils.MASTER_ACCOUNT_ARN_FORMAT.format(
            partition, org["MasterAccountId"], org["Id"]
        )
    )
    assert org["Arn"] == (
        utils.ORGANIZATION_ARN_FORMAT.format(
            partition, org["MasterAccountId"], org["Id"]
        )
    )
    assert org["MasterAccountEmail"] == utils.MASTER_ACCOUNT_EMAIL
    assert org["FeatureSet"] in ["ALL", "CONSOLIDATED_BILLING"]
    assert org["AvailablePolicyTypes"] == [
        {"Type": "SERVICE_CONTROL_POLICY", "Status": "ENABLED"}
    ]


def validate_roots(org, response):
    assert isinstance(response["Roots"], list)
    assert response["Roots"] != []
    root = response["Roots"][0]
    assert re.match(utils.ROOT_ID_REGEX, root["Id"])
    assert root["Arn"] == (
        utils.ROOT_ARN_FORMAT.format(
            "aws", org["MasterAccountId"], org["Id"], root["Id"]
        )
    )
    assert isinstance(root["Name"], str)
    assert root["PolicyTypes"] == []


def validate_organizational_unit(org, response):
    assert isinstance(response["OrganizationalUnit"], dict)
    ou = response["OrganizationalUnit"]
    assert re.match(utils.OU_ID_REGEX, ou["Id"])
    assert ou["Arn"] == (
        utils.OU_ARN_FORMAT.format("aws", org["MasterAccountId"], org["Id"], ou["Id"])
    )
    assert isinstance(ou["Name"], str)


def validate_account(org, account):
    assert sorted(account.keys()) == [
        "Arn",
        "Email",
        "Id",
        "JoinedMethod",
        "JoinedTimestamp",
        "Name",
        "Status",
    ]
    assert re.match(utils.ACCOUNT_ID_REGEX, account["Id"])
    assert account["Arn"] == (
        utils.ACCOUNT_ARN_FORMAT.format(
            "aws", org["MasterAccountId"], org["Id"], account["Id"]
        )
    )
    assert re.match(utils.EMAIL_REGEX, account["Email"])
    assert account["JoinedMethod"] in ["INVITED", "CREATED"]
    assert account["Status"] in ["ACTIVE", "SUSPENDED"]
    assert isinstance(account["Name"], str)
    assert isinstance(account["JoinedTimestamp"], datetime.datetime)


def validate_create_account_status(create_status):
    assert sorted(create_status.keys()) == [
        "AccountId",
        "AccountName",
        "CompletedTimestamp",
        "Id",
        "RequestedTimestamp",
        "State",
    ]
    assert re.match(utils.CREATE_ACCOUNT_STATUS_ID_REGEX, create_status["Id"])
    assert re.match(utils.ACCOUNT_ID_REGEX, create_status["AccountId"])
    assert isinstance(create_status["AccountName"], str)
    assert create_status["State"] == "SUCCEEDED"
    assert isinstance(create_status["RequestedTimestamp"], datetime.datetime)
    assert isinstance(create_status["CompletedTimestamp"], datetime.datetime)


def validate_policy_summary(org, summary):
    assert isinstance(summary, dict)
    assert re.match(utils.POLICY_ID_REGEX, summary["Id"])
    assert summary["Arn"] == (
        utils.SCP_ARN_FORMAT.format(
            "aws", org["MasterAccountId"], org["Id"], summary["Id"]
        )
    )
    assert isinstance(summary["Name"], str)
    assert isinstance(summary["Description"], str)
    assert summary["Type"] == "SERVICE_CONTROL_POLICY"
    assert isinstance(summary["AwsManaged"], bool)


def validate_service_control_policy(org, response):
    assert isinstance(response["PolicySummary"], dict)
    assert isinstance(response["Content"], str)
    validate_policy_summary(org, response["PolicySummary"])



from moto.core.exceptions import JsonRESTError


class NotFoundException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("NotFoundException", message)


class ValidationException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("ValidationException", message)


class AlreadyExistsException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("AlreadyExistsException", message)


class NotAuthorizedException(JsonRESTError):
    code = 400

    def __init__(self) -> None:
        super().__init__("NotAuthorizedException", "")

        self.description = '{"__type":"NotAuthorizedException"}'


class AccessDeniedException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__("AccessDeniedException", message)

        self.description = '{"__type":"AccessDeniedException"}'


class InvalidCiphertextException(JsonRESTError):
    code = 400

    def __init__(self) -> None:
        super().__init__("InvalidCiphertextException", "")

        self.description = '{"__type":"InvalidCiphertextException"}'



import datetime
from typing import Final

from moto.stepfunctions.parser.asl.component.state.wait.wait_function.wait_function import (
    WaitFunction,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class Timestamp(WaitFunction):
    # Timestamp
    # An absolute time to state_wait until beginning the state specified in the Next field.
    # Timestamps must conform to the RFC3339 profile of ISO 8601, with the further
    # restrictions that an uppercase T must separate the date and time portions, and
    # an uppercase Z must denote that a numeric time zone offset is not present, for
    # example, 2016-08-18T17:33:00Z.
    # Note
    # Currently, if you specify the state_wait time as a timestamp, Step Functions considers
    # the time value up to seconds and truncates milliseconds.

    TIMESTAMP_FORMAT: Final[str] = "%Y-%m-%dT%H:%M:%SZ"

    def __init__(self, timestamp):
        self.timestamp: Final[datetime.datetime] = timestamp

    @staticmethod
    def parse_timestamp(timestamp: str) -> datetime.datetime:
        return datetime.datetime.strptime(timestamp, Timestamp.TIMESTAMP_FORMAT)

    def _get_wait_seconds(self, env: Environment) -> int:
        delta = self.timestamp - datetime.datetime.today()
        delta_sec = int(delta.total_seconds())
        return delta_sec



import os
from functools import wraps
from unittest import SkipTest

import boto3
import botocore

from moto import mock_aws


def sns_aws_verified(func):
    """
    Function that is verified to work against AWS.
    Can be run against AWS at any time by setting:
      MOTO_TEST_ALLOW_AWS_REQUEST=true

    If this environment variable is not set, the function runs in a `mock_aws` context.
    """

    @wraps(func)
    def pagination_wrapper():
        allow_aws_request = (
            os.environ.get("MOTO_TEST_ALLOW_AWS_REQUEST", "false").lower() == "true"
        )

        if allow_aws_request:
            ssm = boto3.client("ssm", "us-east-1")
            try:
                param = ssm.get_parameter(
                    Name="/moto/tests/ses/firebase_api_key", WithDecryption=True
                )
                api_key = param["Parameter"]["Value"]
                resp = func(api_key)
            except botocore.exceptions.ClientError:
                # SNS tests try to create a PlatformApplication that connects to GCM
                # (Google Cloud Messaging, also known as Firebase Messaging)
                # That requires an account and an API key
                # If the API key has not been configured in SSM, we'll just skip the test
                #
                # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sns/client/create_platform_application.html
                # AWS calls it 'API key', but Firebase calls it Server Key
                #
                # https://stackoverflow.com/a/75896532/13245310
                raise SkipTest("Can't execute SNS tests without Firebase API key")
        else:
            with mock_aws():
                resp = func("mock_api_key")
        return resp

    return pagination_wrapper



"""Unit tests for networkmanager-supported APIs."""

import boto3

from moto import mock_aws
from tests import DEFAULT_ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


def create_global_network(client) -> str:
    return client.create_global_network(
        Description="Test global network",
        Tags=[
            {"Key": "Name", "Value": "TestNetwork"},
        ],
    )["GlobalNetwork"]["GlobalNetworkId"]


@mock_aws
def test_create_global_network():
    client = boto3.client("networkmanager")
    resp = client.create_global_network(
        Description="Test global network",
        Tags=[
            {"Key": "Name", "Value": "TestNetwork"},
        ],
    )

    global_network = resp["GlobalNetwork"]
    assert (
        global_network["GlobalNetworkArn"]
        == f"arn:aws:networkmanager:{DEFAULT_ACCOUNT_ID}:global-network/{global_network['GlobalNetworkId']}"
    )
    assert global_network["Description"] == "Test global network"
    assert global_network["Tags"] == [{"Key": "Name", "Value": "TestNetwork"}]
    assert global_network["State"] == "PENDING"


@mock_aws
def test_create_core_network():
    client = boto3.client("networkmanager")
    # Create a global network
    global_network_id = client.create_global_network(
        Description="Test global network",
        Tags=[
            {"Key": "Name", "Value": "TestNetwork"},
        ],
    )["GlobalNetwork"]["GlobalNetworkId"]

    resp = client.create_core_network(
        GlobalNetworkId=global_network_id,
        Description="Test core network",
        Tags=[
            {"Key": "Name", "Value": "TestNetwork"},
        ],
        PolicyDocument="policy-document",
        ClientToken="client-token",
    )

    core_network = resp["CoreNetwork"]
    assert (
        core_network["CoreNetworkArn"]
        == f"arn:aws:networkmanager:{DEFAULT_ACCOUNT_ID}:core-network/{core_network['CoreNetworkId']}"
    )
    assert core_network["GlobalNetworkId"] == global_network_id
    assert core_network["Description"] == "Test core network"
    assert len(core_network["Tags"]) == 1


@mock_aws
def test_delete_core_network():
    client = boto3.client("networkmanager")
    gn_id = create_global_network(client)
    core_network = client.create_core_network(GlobalNetworkId=gn_id)
    cn_id = core_network["CoreNetwork"]["CoreNetworkId"]
    assert len(client.list_core_networks()["CoreNetworks"]) == 1
    resp = client.delete_core_network(CoreNetworkId=cn_id)
    assert resp["CoreNetwork"]["CoreNetworkId"] == cn_id
    assert resp["CoreNetwork"]["State"] == "DELETING"
    assert len(client.list_core_networks()["CoreNetworks"]) == 0


@mock_aws
def test_tag_resource():
    client = boto3.client("networkmanager")
    gn_id = create_global_network(client)
    cn = client.create_core_network(GlobalNetworkId=gn_id)["CoreNetwork"]

    # Check tagging core-network
    resp = client.tag_resource(
        ResourceArn=cn["CoreNetworkArn"],
        Tags=[{"Key": "Test", "Value": "TestValue-Core"}],
    )
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    updated_cn = client.get_core_network(CoreNetworkId=cn["CoreNetworkId"])[
        "CoreNetwork"
    ]
    assert updated_cn["Tags"] == [{"Key": "Test", "Value": "TestValue-Core"}]

    # Check tagging global-network
    gn_arn = client.describe_global_networks()["GlobalNetworks"][0]["GlobalNetworkArn"]
    resp = client.tag_resource(
        ResourceArn=gn_arn, Tags=[{"Key": "Test", "Value": "TestValue-Global"}]
    )
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    updated_gn = client.describe_global_networks(GlobalNetworkIds=[gn_id])[
        "GlobalNetworks"
    ][0]
    assert len(updated_gn["Tags"]) == 2
    assert updated_gn["Tags"] == [
        {"Key": "Name", "Value": "TestNetwork"},
        {"Key": "Test", "Value": "TestValue-Global"},
    ]


@mock_aws
def test_untag_resource():
    client = boto3.client("networkmanager")
    gn_id = create_global_network(client)
    cn = client.create_core_network(
        GlobalNetworkId=gn_id,
        Tags=[
            {"Key": "Name", "Value": "TestNetwork"},
            {"Key": "DeleteMe", "Value": "DeleteThisTag!"},
        ],
    )["CoreNetwork"]

    # Check untagging core-network
    resp = client.untag_resource(ResourceArn=cn["CoreNetworkArn"], TagKeys=["DeleteMe"])
    assert resp["ResponseMetadata"]["HTTPStatusCode"] == 200
    updated_cn = client.get_core_network(CoreNetworkId=cn["CoreNetworkId"])[
        "CoreNetwork"
    ]
    assert len(updated_cn["Tags"]) == 1
    assert updated_cn["Tags"] == [{"Key": "Name", "Value": "TestNetwork"}]


@mock_aws
def test_list_core_networks():
    NUM_CORE_NETWORKS = 3
    client = boto3.client("networkmanager")
    for _ in range(NUM_CORE_NETWORKS):
        gn_id = create_global_network(client)
        client.create_core_network(GlobalNetworkId=gn_id)

    resp = client.list_core_networks()
    assert len(resp["CoreNetworks"]) == NUM_CORE_NETWORKS


@mock_aws
def test_get_core_network():
    client = boto3.client("networkmanager")
    gn_id = create_global_network(client)
    cn_id = client.create_core_network(
        GlobalNetworkId=gn_id,
        Description="Test core network",
        Tags=[
            {"Key": "Name", "Value": "TestNetwork"},
        ],
        PolicyDocument="policy-document",
        ClientToken="client-token",
    )["CoreNetwork"]["CoreNetworkId"]

    resp = client.get_core_network(CoreNetworkId=cn_id)
    assert resp["CoreNetwork"]["CoreNetworkId"] == cn_id
    assert resp["CoreNetwork"]["Description"] == "Test core network"
    assert len(resp["CoreNetwork"]["Tags"]) == 1


@mock_aws
def test_describe_global_networks():
    NUM_NETWORKS = 3
    client = boto3.client("networkmanager")
    global_ids = []
    for i in range(NUM_NETWORKS):
        global_id = client.create_global_network(
            Description=f"Test global network #{i}",
            Tags=[
                {"Key": "Name", "Value": f"TestNetwork-{i}"},
            ],
        )["GlobalNetwork"]["GlobalNetworkId"]
        global_ids.append(global_id)
    resp = client.describe_global_networks()
    assert len(resp["GlobalNetworks"]) == NUM_NETWORKS

    # Check each global network by ID
    for g_id in global_ids:
        gn = client.describe_global_networks(GlobalNetworkIds=[g_id])["GlobalNetworks"][
            0
        ]
        assert gn["GlobalNetworkId"] == g_id



from operator import itemgetter

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.cloudwatch.utils import make_arn_for_alarm
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID


@mock_aws
def test_list_tags_for_resource():
    # given
    client = boto3.client("cloudwatch", region_name="eu-central-1")
    alarm_name = "test-alarm"
    client.put_metric_alarm(
        AlarmName=alarm_name,
        AlarmDescription="test alarm",
        ActionsEnabled=True,
        MetricName="5XXError",
        Namespace="AWS/ApiGateway",
        Statistic="Sum",
        Dimensions=[
            {"Name": "ApiName", "Value": "test-api"},
            {"Name": "Stage", "Value": "default"},
        ],
        Period=60,
        Unit="Seconds",
        EvaluationPeriods=1,
        DatapointsToAlarm=1,
        Threshold=1.0,
        ComparisonOperator="GreaterThanOrEqualToThreshold",
        Tags=[{"Key": "key-1", "Value": "value-1"}],
    )
    arn = client.describe_alarms(AlarmNames=[alarm_name])["MetricAlarms"][0]["AlarmArn"]

    # when
    response = client.list_tags_for_resource(ResourceARN=arn)

    # then
    assert response["Tags"] == [{"Key": "key-1", "Value": "value-1"}]


@mock_aws
def test_list_tags_for_resource_with_unknown_resource():
    # given
    region_name = "eu-central-1"
    client = boto3.client("cloudwatch", region_name=region_name)

    # when
    response = client.list_tags_for_resource(
        ResourceARN=make_arn_for_alarm(
            region=region_name, account_id=ACCOUNT_ID, alarm_name="unknown"
        )
    )

    # then
    assert response["Tags"] == []


@mock_aws
def test_tag_resource():
    # given
    client = boto3.client("cloudwatch", region_name="eu-central-1")
    alarm_name = "test-alarm"
    client.put_metric_alarm(
        AlarmName=alarm_name,
        AlarmDescription="test alarm",
        ActionsEnabled=True,
        MetricName="5XXError",
        Namespace="AWS/ApiGateway",
        Statistic="Sum",
        Dimensions=[
            {"Name": "ApiName", "Value": "test-api"},
            {"Name": "Stage", "Value": "default"},
        ],
        Period=60,
        Unit="Seconds",
        EvaluationPeriods=1,
        DatapointsToAlarm=1,
        Threshold=1.0,
        ComparisonOperator="GreaterThanOrEqualToThreshold",
        Tags=[{"Key": "key-1", "Value": "value-1"}],
    )
    arn = client.describe_alarms(AlarmNames=[alarm_name])["MetricAlarms"][0]["AlarmArn"]

    # when
    client.tag_resource(ResourceARN=arn, Tags=[{"Key": "key-2", "Value": "value-2"}])

    # then
    response = client.list_tags_for_resource(ResourceARN=arn)
    assert sorted(response["Tags"], key=itemgetter("Key")) == [
        {"Key": "key-1", "Value": "value-1"},
        {"Key": "key-2", "Value": "value-2"},
    ]


@mock_aws
def test_tag_resource_on_resource_without_tags():
    cw = boto3.client("cloudwatch", region_name="eu-central-1")
    cw.put_metric_alarm(
        AlarmName="testalarm",
        EvaluationPeriods=1,
        ComparisonOperator="GreaterThanThreshold",
        Period=60,
        MetricName="test",
        Namespace="test",
    )
    alarms = cw.describe_alarms()
    alarm_arn = alarms["MetricAlarms"][0]["AlarmArn"]

    # List 0 tags - none have been added
    assert cw.list_tags_for_resource(ResourceARN=alarm_arn)["Tags"] == []

    # Tag the Alarm for the first time
    cw.tag_resource(ResourceARN=alarm_arn, Tags=[{"Key": "tk", "Value": "tv"}])
    assert len(cw.list_tags_for_resource(ResourceARN=alarm_arn)["Tags"]) == 1


@mock_aws
def test_tag_resource_error_not_exists():
    # given
    region_name = "eu-central-1"
    client = boto3.client("cloudwatch", region_name=region_name)

    # when
    with pytest.raises(ClientError) as e:
        client.tag_resource(
            ResourceARN=make_arn_for_alarm(
                region=region_name, account_id=ACCOUNT_ID, alarm_name="unknown"
            ),
            Tags=[{"Key": "key-1", "Value": "value-1"}],
        )

    # then
    ex = e.value
    assert ex.operation_name == "TagResource"
    assert ex.response["ResponseMetadata"]["HTTPStatusCode"] == 404
    assert ex.response["Error"]["Code"] == "ResourceNotFoundException"
    assert ex.response["Error"]["Message"] == "Unknown"


@mock_aws
def test_untag_resource():
    # given
    client = boto3.client("cloudwatch", region_name="eu-central-1")
    alarm_name = "test-alarm"
    client.put_metric_alarm(
        AlarmName=alarm_name,
        AlarmDescription="test alarm",
        ActionsEnabled=True,
        MetricName="5XXError",
        Namespace="AWS/ApiGateway",
        Statistic="Sum",
        Dimensions=[
            {"Name": "ApiName", "Value": "test-api"},
            {"Name": "Stage", "Value": "default"},
        ],
        Period=60,
        Unit="Seconds",
        EvaluationPeriods=1,
        DatapointsToAlarm=1,
        Threshold=1.0,
        ComparisonOperator="GreaterThanOrEqualToThreshold",
        Tags=[
            {"Key": "key-1", "Value": "value-1"},
            {"Key": "key-2", "Value": "value-2"},
        ],
    )
    arn = client.describe_alarms(AlarmNames=[alarm_name])["MetricAlarms"][0]["AlarmArn"]

    # when
    client.untag_resource(ResourceARN=arn, TagKeys=["key-2"])

    # then
    response = client.list_tags_for_resource(ResourceARN=arn)
    assert response["Tags"] == [{"Key": "key-1", "Value": "value-1"}]


@mock_aws
def test_untag_resource_error_not_exists():
    # given
    region_name = "eu-central-1"
    client = boto3.client("cloudwatch", region_name=region_name)

    # when
    with pytest.raises(ClientError) as e:
        client.untag_resource(
            ResourceARN=make_arn_for_alarm(
                region=region_name, account_id=ACCOUNT_ID, alarm_name="unknown"
            ),
            TagKeys=["key-1"],
        )

    # then
    ex = e.value
    assert ex.operation_name == "UntagResource"
    assert ex.response["ResponseMetadata"]["HTTPStatusCode"] == 404
    assert ex.response["Error"]["Code"] == "ResourceNotFoundException"
    assert ex.response["Error"]["Message"] == "Unknown"



import json
from typing import Any, Dict

from moto.core.responses import BaseResponse
from moto.route53domains.models import Route53DomainsBackend, route53domains_backends
from moto.route53domains.validators import Route53Domain, Route53DomainsOperation


class Route53DomainsResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="route53-domains")

    @property
    def route53domains_backend(self) -> Route53DomainsBackend:
        return route53domains_backends[self.current_account][self.partition]

    def register_domain(self) -> str:
        domain_name = self._get_param("DomainName")
        duration_in_years = self._get_int_param("DurationInYears")
        auto_renew = self._get_bool_param("AutoRenew", if_none=True)
        admin_contact = self._get_param("AdminContact")
        registrant_contact = self._get_param("RegistrantContact")
        tech_contact = self._get_param("TechContact")
        privacy_protection_admin_contact = self._get_bool_param(
            "PrivacyProtectAdminContact", if_none=True
        )
        privacy_protection_registrant_contact = self._get_bool_param(
            "PrivacyProtectRegistrantContact", if_none=True
        )
        privacy_protection_tech_contact = self._get_bool_param(
            "PrivacyProtectTechContact", if_none=True
        )
        extra_params = self._get_param("ExtraParams")

        operation = self.route53domains_backend.register_domain(
            domain_name=domain_name,
            duration_in_years=duration_in_years,
            auto_renew=auto_renew,
            admin_contact=admin_contact,
            registrant_contact=registrant_contact,
            tech_contact=tech_contact,
            private_protect_admin_contact=privacy_protection_admin_contact,
            private_protect_registrant_contact=privacy_protection_registrant_contact,
            private_protect_tech_contact=privacy_protection_tech_contact,
            extra_params=extra_params,
        )

        return json.dumps({"OperationId": operation.id})

    def delete_domain(self) -> str:
        domain_name = self._get_param("DomainName")
        operation = self.route53domains_backend.delete_domain(domain_name)

        return json.dumps({"OperationId": operation.id})

    def get_domain_detail(self) -> str:
        domain_name = self._get_param("DomainName")

        return json.dumps(
            self.route53domains_backend.get_domain(domain_name=domain_name).to_json()
        )

    def list_domains(self) -> str:
        filter_conditions = self._get_param("FilterConditions")
        sort_condition = self._get_param("SortCondition")
        marker = self._get_param("Marker")
        max_items = self._get_param("MaxItems")
        domains, marker = self.route53domains_backend.list_domains(
            filter_conditions=filter_conditions,
            sort_condition=sort_condition,
            marker=marker,
            max_items=max_items,
        )
        res = {
            "Domains": list(map(self.__map_domains_to_info, domains)),
        }

        if marker:
            res["NextPageMarker"] = marker

        return json.dumps(res)

    @staticmethod
    def __map_domains_to_info(domain: Route53Domain) -> Dict[str, Any]:  # type: ignore[misc]
        return {
            "DomainName": domain.domain_name,
            "AutoRenew": domain.auto_renew,
            "Expiry": domain.expiration_date.timestamp(),
            "TransferLock": True,
        }

    def list_operations(self) -> str:
        submitted_since_timestamp = self._get_int_param("SubmittedSince")
        max_items = self._get_int_param("MaxItems")
        statuses = self._get_param("Status")
        marker = self._get_param("Marker")
        types = self._get_param("Type")
        sort_by = self._get_param("SortBy")
        sort_order = self._get_param("SortOrder")

        operations, marker = self.route53domains_backend.list_operations(
            submitted_since_timestamp=submitted_since_timestamp,
            max_items=max_items,
            marker=marker,
            statuses=statuses,
            types=types,
            sort_by=sort_by,
            sort_order=sort_order,
        )

        res = {
            "Operations": [operation.to_json() for operation in operations],
        }

        if marker:
            res["NextPageMarker"] = marker

        return json.dumps(res)

    def get_operation_detail(self) -> str:
        operation_id = self._get_param("OperationId")
        operation: Route53DomainsOperation = self.route53domains_backend.get_operation(
            operation_id
        )

        return json.dumps(operation.to_json())

    def update_domain_nameservers(self) -> str:
        domain_name = self._get_param("DomainName")
        nameservers = self._get_param("Nameservers")

        operation: Route53DomainsOperation = (
            self.route53domains_backend.update_domain_nameservers(
                domain_name=domain_name, nameservers=nameservers
            )
        )

        return json.dumps({"OperationId": operation.id})



"""Handles incoming sagemakermetrics requests, invokes methods, returns responses."""

import json

from moto.core.responses import BaseResponse

from .models import SageMakerMetricsBackend, sagemakermetrics_backends


class SageMakerMetricsResponse(BaseResponse):
    """Handler for SageMakerMetrics requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="sagemaker-metrics")

    @property
    def sagemakermetrics_backend(self) -> SageMakerMetricsBackend:
        """Return backend instance specific for this region."""
        return sagemakermetrics_backends[self.current_account][self.region]

    def batch_put_metrics(self) -> str:
        trial_component_name = self._get_param("TrialComponentName")
        metric_data = self._get_param("MetricData")
        errors = self.sagemakermetrics_backend.batch_put_metrics(
            trial_component_name=trial_component_name,
            metric_data=metric_data,
        )
        return json.dumps(errors)



# Standard library modules
import unittest

# Third-party modules
import boto3
from botocore.exceptions import ClientError

# Package modules
from moto import mock_aws

AWS_REGION = "us-west-1"

SG_STACK_NAME = "simple-sg-stack"
SG_TEMPLATE = """
AWSTemplateFormatVersion: 2010-09-09
Description: Simple test CF template for moto_cloudformation


Resources:
  SimpleSecurityGroup:
    Type: AWS::EC2::SecurityGroup
    Description: "A simple security group"
    Properties:
      GroupName: simple-security-group
      GroupDescription: "A simple security group"
      SecurityGroupEgress:
        -
          Description: "Egress to remote HTTPS servers"
          CidrIp: 0.0.0.0/0
          IpProtocol: tcp
          FromPort: 443
          ToPort: 443

Outputs:
    SimpleSecurityGroupName:
        Value: !GetAtt SimpleSecurityGroup.GroupId
        Export:
            Name: "SimpleSecurityGroup"

"""

EC2_STACK_NAME = "simple-ec2-stack"
EC2_TEMPLATE = """
---
# The latest template format version is "2010-09-09" and as of 2018-04-09
# is currently the only valid value.
AWSTemplateFormatVersion: 2010-09-09
Description: Simple test CF template for moto_cloudformation


Resources:
  SimpleInstance:
    Type: AWS::EC2::Instance
    Properties:
        ImageId: ami-03cf127a
        InstanceType: t2.micro
        SecurityGroups: !Split [',', !ImportValue SimpleSecurityGroup]
"""


class TestSimpleInstance(unittest.TestCase):
    def test_simple_instance(self):
        """Test that we can create a simple CloudFormation stack that imports values from an existing CloudFormation stack"""
        with mock_aws():
            client = boto3.client("cloudformation", region_name=AWS_REGION)
            client.create_stack(StackName=SG_STACK_NAME, TemplateBody=SG_TEMPLATE)
            response = client.create_stack(
                StackName=EC2_STACK_NAME, TemplateBody=EC2_TEMPLATE
            )
            self.assertIn("StackId", response)
            response = client.describe_stacks(StackName=response["StackId"])
            self.assertIn("Stacks", response)
            stack_info = response["Stacks"]
            self.assertEqual(1, len(stack_info))
            self.assertIn("StackName", stack_info[0])
            self.assertEqual(EC2_STACK_NAME, stack_info[0]["StackName"])

    def test_simple_instance_missing_export(self):
        """Test that we get an exception if a CloudFormation stack tries to imports a non-existent export value"""
        with mock_aws():
            client = boto3.client("cloudformation", region_name=AWS_REGION)
            with self.assertRaises(ClientError) as e:
                client.create_stack(StackName=EC2_STACK_NAME, TemplateBody=EC2_TEMPLATE)
            self.assertIn("Error", e.exception.response)
            self.assertIn("Code", e.exception.response["Error"])
            self.assertEqual("ExportNotFound", e.exception.response["Error"]["Code"])



import sys
from importlib.metadata import version

from moto.utilities.distutils_version import LooseVersion

PYTHON_VERSION_INFO = sys.version_info
PYTHON_311 = sys.version_info >= (3, 11)
RESPONSES_VERSION = version("responses")
WERKZEUG_VERSION = version("werkzeug")


def is_responses_0_17_x() -> bool:
    return LooseVersion(RESPONSES_VERSION) >= LooseVersion("0.17.0")


def is_werkzeug_2_0_x_or_older() -> bool:
    return LooseVersion(WERKZEUG_VERSION) < LooseVersion("2.1.0")


def is_werkzeug_2_3_x() -> bool:
    return LooseVersion(WERKZEUG_VERSION) >= LooseVersion("2.3.0")



import boto3
import freezegun
import pytest
from botocore.client import ClientError

from moto import mock_aws
from moto.settings import TEST_SERVER_MODE


@freezegun.freeze_time("2022-06-01 12:00:00")
@mock_aws
def test_create_function_definition():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    init_ver = {
        "Functions": [
            {
                "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
                "Id": "1234567890",
                "FunctionConfiguration": {
                    "MemorySize": 16384,
                    "EncodingType": "binary",
                    "Pinned": True,
                    "Timeout": 3,
                },
            }
        ]
    }
    func_name = "TestFunc"
    res = client.create_function_definition(InitialVersion=init_ver, Name=func_name)
    assert "Arn" in res
    assert "Id" in res
    assert "LatestVersion" in res
    assert "LatestVersionArn" in res
    assert res["Name"] == func_name
    assert res["ResponseMetadata"]["HTTPStatusCode"] == 201

    if not TEST_SERVER_MODE:
        assert res["CreationTimestamp"] == "2022-06-01T12:00:00.000Z"
        assert res["LastUpdatedTimestamp"] == "2022-06-01T12:00:00.000Z"


@freezegun.freeze_time("2022-06-01 12:00:00")
@mock_aws
def test_list_function_definitions():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    init_ver = {
        "Functions": [
            {
                "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
                "Id": "1234567890",
                "FunctionConfiguration": {
                    "MemorySize": 16384,
                    "EncodingType": "binary",
                    "Pinned": True,
                    "Timeout": 3,
                },
            }
        ]
    }
    func_name = "TestFunc"
    client.create_function_definition(InitialVersion=init_ver, Name=func_name)

    res = client.list_function_definitions()
    func_def = res["Definitions"][0]

    assert func_def["Name"] == func_name
    assert "Arn" in func_def
    assert "Id" in func_def
    assert "LatestVersion" in func_def
    assert "LatestVersionArn" in func_def
    if not TEST_SERVER_MODE:
        assert func_def["CreationTimestamp"] == "2022-06-01T12:00:00.000Z"
        assert func_def["LastUpdatedTimestamp"] == "2022-06-01T12:00:00.000Z"


@freezegun.freeze_time("2022-06-01 12:00:00")
@mock_aws
def test_get_function_definition():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    init_ver = {
        "Functions": [
            {
                "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
                "Id": "1234567890",
                "FunctionConfiguration": {
                    "MemorySize": 16384,
                    "EncodingType": "binary",
                    "Pinned": True,
                    "Timeout": 3,
                },
            }
        ]
    }
    func_name = "TestFunc"
    create_res = client.create_function_definition(
        InitialVersion=init_ver, Name=func_name
    )

    func_def_id = create_res["Id"]
    arn = create_res["Arn"]
    latest_version = create_res["LatestVersion"]
    latest_version_arn = create_res["LatestVersionArn"]

    get_res = client.get_function_definition(FunctionDefinitionId=func_def_id)

    assert get_res["Name"] == func_name
    assert get_res["Arn"] == arn
    assert get_res["Id"] == func_def_id
    assert get_res["LatestVersion"] == latest_version
    assert get_res["LatestVersionArn"] == latest_version_arn
    assert get_res["ResponseMetadata"]["HTTPStatusCode"] == 200

    if not TEST_SERVER_MODE:
        assert get_res["CreationTimestamp"] == "2022-06-01T12:00:00.000Z"
        assert get_res["LastUpdatedTimestamp"] == "2022-06-01T12:00:00.000Z"


@mock_aws
def test_get_function_definition_with_invalid_id():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    with pytest.raises(ClientError) as ex:
        client.get_function_definition(
            FunctionDefinitionId="b552443b-1888-469b-81f8-0ebc5ca92949"
        )
    err = ex.value.response["Error"]
    assert err["Message"] == "That Lambda List Definition does not exist."
    assert err["Code"] == "IdNotFoundException"


@mock_aws
def test_delete_function_definition():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    init_ver = {
        "Functions": [
            {
                "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
                "Id": "1234567890",
                "FunctionConfiguration": {
                    "MemorySize": 16384,
                    "EncodingType": "binary",
                    "Pinned": True,
                    "Timeout": 3,
                },
            }
        ]
    }
    create_res = client.create_function_definition(
        InitialVersion=init_ver, Name="TestFunc"
    )

    func_def_id = create_res["Id"]
    del_res = client.delete_function_definition(FunctionDefinitionId=func_def_id)
    assert del_res["ResponseMetadata"]["HTTPStatusCode"] == 200


@mock_aws
def test_delete_function_definition_with_invalid_id():
    client = boto3.client("greengrass", region_name="ap-northeast-1")

    with pytest.raises(ClientError) as ex:
        client.delete_function_definition(
            FunctionDefinitionId="6fbffc21-989e-4d29-a793-a42f450a78c6"
        )
    err = ex.value.response["Error"]
    assert err["Message"] == "That lambdas definition does not exist."
    assert err["Code"] == "IdNotFoundException"


@mock_aws
def test_update_function_definition():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    init_ver = {
        "Functions": [
            {
                "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
                "Id": "1234567890",
                "FunctionConfiguration": {
                    "MemorySize": 16384,
                    "EncodingType": "binary",
                    "Pinned": True,
                    "Timeout": 3,
                },
            }
        ]
    }
    create_res = client.create_function_definition(
        InitialVersion=init_ver, Name="TestFunc"
    )

    func_def_id = create_res["Id"]
    updated_func_name = "UpdatedFunction"
    update_res = client.update_function_definition(
        FunctionDefinitionId=func_def_id, Name=updated_func_name
    )
    assert update_res["ResponseMetadata"]["HTTPStatusCode"] == 200

    get_res = client.get_function_definition(FunctionDefinitionId=func_def_id)
    assert get_res["Name"] == updated_func_name


@mock_aws
def test_update_function_definition_with_empty_name():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    init_ver = {
        "Functions": [
            {
                "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
                "Id": "1234567890",
                "FunctionConfiguration": {
                    "MemorySize": 16384,
                    "EncodingType": "binary",
                    "Pinned": True,
                    "Timeout": 3,
                },
            }
        ]
    }
    create_res = client.create_function_definition(
        InitialVersion=init_ver, Name="TestFunc"
    )

    func_def_id = create_res["Id"]

    with pytest.raises(ClientError) as ex:
        client.update_function_definition(FunctionDefinitionId=func_def_id, Name="")
    err = ex.value.response["Error"]
    assert err["Message"] == "Input does not contain any attributes to be updated"
    assert err["Code"] == "InvalidContainerDefinitionException"


@mock_aws
def test_update_function_definition_with_invalid_id():
    client = boto3.client("greengrass", region_name="ap-northeast-1")

    with pytest.raises(ClientError) as ex:
        client.update_function_definition(
            FunctionDefinitionId="6fbffc21-989e-4d29-a793-a42f450a78c6", Name="123"
        )
    rr = ex.value.response["Error"]
    assert rr["Message"] == "That lambdas definition does not exist."
    assert rr["Code"] == "IdNotFoundException"


@freezegun.freeze_time("2022-06-01 12:00:00")
@mock_aws
def test_create_function_definition_version():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    v1_functions = [
        {
            "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
            "Id": "1234567890",
            "FunctionConfiguration": {
                "MemorySize": 16384,
                "EncodingType": "binary",
                "Pinned": True,
                "Timeout": 3,
            },
        }
    ]

    initial_version = {"Functions": v1_functions}
    func_def_res = client.create_function_definition(
        InitialVersion=initial_version, Name="TestFunction"
    )
    func_def_id = func_def_res["Id"]

    v2_functions = [
        {
            "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:2",
            "Id": "987654321",
            "FunctionConfiguration": {
                "MemorySize": 128,
                "EncodingType": "binary",
                "Pinned": False,
                "Timeout": 5,
            },
        }
    ]

    func_def_ver_res = client.create_function_definition_version(
        FunctionDefinitionId=func_def_id, Functions=v2_functions
    )
    assert "Arn" in func_def_ver_res
    assert "CreationTimestamp" in func_def_ver_res
    assert func_def_ver_res["Id"] == func_def_id
    assert "Version" in func_def_ver_res

    if not TEST_SERVER_MODE:
        assert func_def_ver_res["CreationTimestamp"] == "2022-06-01T12:00:00.000Z"


@mock_aws
def test_create_function_definition_version_with_invalid_id():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    functions = [
        {
            "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
            "Id": "1234567890",
            "FunctionConfiguration": {
                "MemorySize": 16384,
                "EncodingType": "binary",
                "Pinned": True,
                "Timeout": 3,
            },
        }
    ]
    with pytest.raises(ClientError) as ex:
        client.create_function_definition_version(
            FunctionDefinitionId="7b0bdeae-54c7-47cf-9f93-561e672efd9c",
            Functions=functions,
        )
    err = ex.value.response["Error"]
    assert err["Message"] == "That lambdas does not exist."
    assert err["Code"] == "IdNotFoundException"


@freezegun.freeze_time("2022-06-01 12:00:00")
@mock_aws
def test_list_function_definition_versions():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    functions = [
        {
            "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
            "Id": "1234567890",
            "FunctionConfiguration": {
                "MemorySize": 16384,
                "EncodingType": "binary",
                "Pinned": True,
                "Timeout": 3,
            },
        }
    ]

    initial_version = {"Functions": functions}
    create_res = client.create_function_definition(
        InitialVersion=initial_version, Name="TestFunction"
    )
    func_def_id = create_res["Id"]

    func_def_vers_res = client.list_function_definition_versions(
        FunctionDefinitionId=func_def_id
    )

    assert "Versions" in func_def_vers_res
    device_def_ver = func_def_vers_res["Versions"][0]
    assert "Arn" in device_def_ver
    assert "CreationTimestamp" in device_def_ver

    if not TEST_SERVER_MODE:
        assert device_def_ver["CreationTimestamp"] == "2022-06-01T12:00:00.000Z"
    assert device_def_ver["Id"] == func_def_id
    assert "Version" in device_def_ver


@mock_aws
def test_list_function_definition_versions_with_invalid_id():
    client = boto3.client("greengrass", region_name="ap-northeast-1")

    with pytest.raises(ClientError) as ex:
        client.list_function_definition_versions(
            FunctionDefinitionId="7b0bdeae-54c7-47cf-9f93-561e672efd9c"
        )
    err = ex.value.response["Error"]
    assert err["Message"] == "That lambdas definition does not exist."
    assert err["Code"] == "IdNotFoundException"


@freezegun.freeze_time("2022-06-01 12:00:00")
@mock_aws
def test_get_function_definition_version():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    functions = [
        {
            "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
            "Id": "1234567890",
            "FunctionConfiguration": {
                "MemorySize": 16384,
                "EncodingType": "binary",
                "Pinned": True,
                "Timeout": 3,
            },
        }
    ]

    initial_version = {"Functions": functions}
    create_res = client.create_function_definition(
        InitialVersion=initial_version, Name="TestFunction"
    )

    func_def_id = create_res["Id"]
    func_def_ver_id = create_res["LatestVersion"]

    func_def_ver_res = client.get_function_definition_version(
        FunctionDefinitionId=func_def_id, FunctionDefinitionVersionId=func_def_ver_id
    )

    assert "Arn" in func_def_ver_res
    assert "CreationTimestamp" in func_def_ver_res
    assert func_def_ver_res["Definition"] == initial_version
    assert func_def_ver_res["Id"] == func_def_id
    assert "Version" in func_def_ver_res

    if not TEST_SERVER_MODE:
        assert func_def_ver_res["CreationTimestamp"] == "2022-06-01T12:00:00.000Z"


@mock_aws
def test_get_function_definition_version_with_invalid_id():
    client = boto3.client("greengrass", region_name="ap-northeast-1")

    with pytest.raises(ClientError) as ex:
        client.get_function_definition_version(
            FunctionDefinitionId="7b0bdeae-54c7-47cf-9f93-561e672efd9c",
            FunctionDefinitionVersionId="7b0bdeae-54c7-47cf-9f93-561e672efd9c",
        )
    err = ex.value.response["Error"]
    assert err["Message"] == "That lambdas definition does not exist."
    assert err["Code"] == "IdNotFoundException"


@mock_aws
def test_get_function_definition_version_with_invalid_version_id():
    client = boto3.client("greengrass", region_name="ap-northeast-1")
    functions = [
        {
            "FunctionArn": "arn:aws:lambda:ap-northeast-1:123456789012:function:test-func:1",
            "Id": "1234567890",
            "FunctionConfiguration": {
                "MemorySize": 16384,
                "EncodingType": "binary",
                "Pinned": True,
                "Timeout": 3,
            },
        }
    ]

    initial_version = {"Functions": functions}
    create_res = client.create_function_definition(
        InitialVersion=initial_version, Name="TestFunction"
    )

    func_def_id = create_res["Id"]
    invalid_func_def_ver_id = "7b0bdeae-54c7-47cf-9f93-561e672efd9c"

    with pytest.raises(ClientError) as ex:
        client.get_function_definition_version(
            FunctionDefinitionId=func_def_id,
            FunctionDefinitionVersionId=invalid_func_def_ver_id,
        )
    err = ex.value.response["Error"]
    assert (
        err["Message"]
        == f"Version {invalid_func_def_ver_id} of Lambda List Definition {func_def_id} does not exist."
    )
    assert err["Code"] == "IdNotFoundException"



"""Exceptions raised by the opensearch service."""

from moto.core.exceptions import JsonRESTError


class ResourceNotFoundException(JsonRESTError):
    def __init__(self, name: str):
        super().__init__("ResourceNotFoundException", f"Domain not found: {name}")


class EngineTypeNotFoundException(JsonRESTError):
    def __init__(self, domain_name: str):
        super().__init__(
            "EngineTypeNotFoundException", f"Engine Type not found: {domain_name}"
        )



import json

import moto.server as server
from moto import mock_aws

"""
Test the different server responses
"""


@mock_aws
def test_medialive_list_channels():
    backend = server.create_backend_app("medialive")
    test_client = backend.test_client()

    res = test_client.get("/prod/channels")
    result = res.data.decode("utf-8")
    assert json.loads(result) == {"channels": [], "nextToken": None}


@mock_aws
def test_medialive_list_inputs():
    backend = server.create_backend_app("medialive")
    test_client = backend.test_client()

    res = test_client.get("/prod/inputs")

    result = res.data.decode("utf-8")
    assert json.loads(result) == {"inputs": [], "nextToken": None}
